(define-module (alterator vm profile)
  :use-module (srfi srfi-1)
  :use-module (srfi srfi-2)
  :use-module (srfi srfi-11)
  :use-module (srfi srfi-13)
  :use-module (alterator vm defs)
  :use-module (alterator vm utils)
  :export (create-profiles))

(define vm-profile-custom
  '(none (title . "Custom") (action . none)))

(define vm-profile-file "/tmp/metadata/vm-profile.scm")
(define vm-profile-file-bundled "/var/cache/alterator/vm-profile.scm")

(define (create-profiles)
  (let ((ext (or (and (access? vm-profile-file R_OK)
                      (map create-profile
                           (with-input-from-file vm-profile-file read)))
                 (and (access? vm-profile-file-bundled R_OK)
                      (map create-profile
                           (with-input-from-file vm-profile-file-bundled read)))
                 '())))
    (append ext (list (create-profile vm-profile-custom)))))

(define (profile-part-size part)
  (assv-ref part 'size))

(define (profile-part-min-size< parta partb)
  (< (pr-min-size (profile-part-size parta))
     (pr-min-size (profile-part-size partb))))

(define (profile-part-exact? part)
  (number? (pr-max-size (profile-part-size part))))

(define (profile-min-size data)
  (apply + (map pr-min-size
                (map profile-part-size
                     (map cdr data)))))

(define (profile-reorder data)
  (let-values (((exact fuzzy) (partition profile-part-exact? (cdr data))))
    (append exact (sort fuzzy profile-part-min-size<))))

(define (profile-size->string profile)
  (if (eqv? (assv-ref  profile 'action) 'none) ""
      (size->string (profile-min-size (assv-ref profile 'actiondata)))))

(define (actiondata-efi actiondata)
  (if (have-efi?)
      (append '(("efi" (size . 524288) (fsim . "FAT32") (methods plain))) actiondata)
      actiondata))

(define (profile-actiondata profile)
  (let ((actiondata (assv-ref profile 'actiondata)))
    (if (any (lambda(disk) (string=? (assumed-segmgr disk) "DosSegMgr"))
             (pr-candidates))
        actiondata
        (append '(("biosboot" (size . 16384) (methods plain))) actiondata))))

(define (create-profile profile)
  (let ((name (symbol->string (car profile)))
        (profile (profile-reorder profile)))
    (lambda (command junk . args)
      (case command
        ((name) name)
        ((value) (lambda () profile))
        ((tree) (list name))
        ((title) `(quote ,(assv-ref profile 'title)))
        ((size) `(quote ,(profile-size->string profile)))
        ((apply)
         (format #t "APPLYING PROFILE: ~S ~S\n" name args)
         (and (assv-ref args 'clearall)
              (pr-clearall (assv-ref args 'commit)))
         (with-fluids
          ((pr-exclude (assv-ref args 'exclude)))
          (case (assv-ref profile 'action)
            ((none) '(quote nexthop /vm/table))
            (else (if (pr-apply (actiondata-efi (profile-actiondata profile)))
                      (if (assv-ref args 'commit)
                          '(quote #t)
                          '(quote nexthop /vm/table))
                      '(quote #f ""))))))))))

;---------------------------------------------------------------
(define pr-exclude (make-fluid))

(define (pr-clearall commit)
  (let ((kill (lambda(x) (and (x 'delete?) (x 'delete)))))
    (let ((killall (lambda(x) (for-each kill (apply evms x)) (and commit (evms 'commit)))))
      (for-each killall
                '((volume-list)
                  (object-list #:type region #:data-type data)
                  (container-list)
                  (object-list #:type region #:data-type data)
                  (object-list #:type segment #:data-type data)))))
  (for-each
   (lambda (x) (and (x 'unassign?) (x 'unassign)))
   (evms 'object-list #:type 'disk))
  (and commit (evms 'commit)))

(define (pr-apply data)
  (every values (unfold-right null? pr-place cdr data '())))

(define (pr-place args)
  (let ((name (caar args))
        (args (cdar args))
        (rest-size (profile-min-size (cdr args))))
    (format #t "PLACE: ~S ~S ~S\n" name args rest-size)
    (pr-make-volume
     name (assv-ref args 'fsim)
     (pr-place-one name
                   (assv-ref args 'size)
                   rest-size
                   (assv-ref args 'methods)))))

(define (pr-make-volume name fsim object)
  (and object
       (format #t "MAKE-VOLUME: ~S ~S ~S\n" name fsim (obj->name (car object)))
       (and-let*
        ((vol (car ((evms 'create-compat (car object)) 'apply))))
        (format #t "MAKE-VOLUME: volume: ~S\n" (obj->name vol))
        (and fsim                       ;PReP doesn't have fsim
             (let ((task (evms 'mkfs (evms 'plugin #:name fsim))))
               (set! (task 'selected) (list vol))
               (task 'apply))
             (format #t "MAKE-VOLUME: fsim: ~S\n" (obj->plugin->name vol))
             ;; do not assign swap mntpoint, will be done later
             (or (equal? (obj->plugin->name vol) "SWAPFS")
                 (let ((task (evms 'setmntent vol)))
                   (or (and (equal? name "efi") (set! (task 'options) `((mntpoint . "/boot/efi"))))
                       (set! (task 'options) `((mntpoint . ,name))))
                   (task 'apply))))
        (format #t "MAKE-VOLUME: finished\n")
        vol)))

(define (pr-place-one name size rest methods)
  (format #t "PLACE-ONE: ~S ~S ~S ~S\n" name size rest methods)
  (if (null? methods) #f
      (or (case (car methods)
            ((lvm) (pr-place-lvm name size rest))
            ((raid) (pr-place-raid name size rest))
            ((plain) (pr-place-plain name size rest))
            (else (format #t "wtf?\n") #f))
          (pr-place-one name size rest (cdr methods)))))

(define (pr-place-size< placea placeb)
  (let-values (((obja junka junktooa) placea)
               ((objb junkb junktoob) placeb))
    (< (obj->size obja) (obj->size objb))))

;---------------------------------------------------------------
(define (pr-place-plain name size rest)
  (let* ((places (pr-list-places size))
         (rest (if (or (null? places) (null? (cdr places))) rest #f)))
    (format #t "PLACE-PLAIN: ~S ~S ~S\n" name size rest)
    (let loop ((places places))
      (and (not (null? places))
           (or (pr-place-plain-one (car places)
                                   (cond ((string=? name "prep") 65)
                                         ((string=? name "swap") 130)
                                         ((string=? name "biosboot") 61186) ; GPT #EF02
                                         ((string=? name "efi") 61184) ; GPT #EF00
                                         (#t 131))
                                   size
                                   rest)
               (loop (cdr places)))))))

(define (pr-list-places size)
  (sort
   (append-map (lambda(dsk) (pr-disk->freespace dsk size))
               (pr-candidates))
   pr-place-size<))

(define (pr-place-size seg size rest)
  (let ((segsize (assv-ref (seg) 'size))
        (minsize (pr-min-size size))
        (maxsize (pr-max-size size)))
    (cond ((number? maxsize) maxsize)
          ((number? rest)
           (round-integer (/ (* minsize segsize) (+ minsize rest))))
          (#t #f))))

(define (pr-place-plain-one place type size rest)
  (format #t "PLACE-PLAIN-ONE: ~S ~S ~S\n" place size rest)
  (false-if-exception
   (let-values (((seg splittable primary) place))
     (let ((task (evms 'create (obj->segmgr seg))))
       (format #t "MAKE-PARTITION: selected: ~S acceptable: ~S\n"
               (obj->name seg) (map obj->name (task 'acceptable)))
       (set! (task 'selected) (list seg))
       (let ((size (pr-place-size seg size rest)))
         (and (number? size)
              (set! (task 'options) `(("Size" . ,size)))))
       (and (eqv? type 65)
            (set! (task 'options) `(("Bootable" . #t))))
       (if (eqv? type 61186)            ; gpt bios-boot type
           (set! (task 'options) `(("Type" . "bios boot partition"))))
       (if (eqv? type 61184)            ; gpt efi type
           (set! (task 'options) `(("Type" . "efi system partition")))
           (set! (task 'options)
                 `(("Primary" . ,primary)
                   ("TypeByNumber" . ,type))))
       (task 'apply)))))

;;--------------------------------------------------------------
(define (pr-lvm-place-pvs total-size)
  (format (current-error-port) "LVM-PLACE-PVS: ~S\n" total-size)
  (let loop ((pvs '())
             (current-size 0)
             (places (pr-list-places 8192)))
    (cond ((< total-size current-size) pvs)
          ((null? places) #f)
          (#t
           (and-let* ((pvv (pr-place-plain-one (car places) 36352 #t #f))
                      (pv (car pvv)))
                     (loop (cons pv pvs)
                           (+ (obj->size pv) current-size)
                           (cdr places)))))))

(define (pr-lvm-place-vg vgname size rest)
  (and-let*
   ((pvs (pr-lvm-place-pvs (+ rest (pr-min-size size))))
    (plugin (evms 'plugin #:name "LVM2"))
    (task (evms 'create-container plugin)))
   (set! (task 'selected) (task 'acceptable))
   (set! (task 'options) `(("name" . ,vgname)))
   (task 'apply)))

(define (pr-lvm-find-vg name)
  (let ((plugin (evms 'plugin #:name "LVM2"))
        (mangled-name (string-append "lvm2|" name)))
    (find
     (lambda(vg) (equal? (obj->name vg) mangled-name))
     (evms 'container-list #:plugin plugin))))

(define (pr-lvm-vg vgname size rest)
  (or (pr-lvm-find-vg vgname)
      (and-let*
       ((vg (pr-lvm-place-vg vgname size rest)))
       (car vg))))

(define (pr-lvm-mangle-lv-name name)
  (cond ((string=? name "/") "root")
        (#t (string-trim name #\/))))

(define (pr-place-lvm name size rest)
  (and-let*
   ((vg (pr-lvm-vg "alt" size rest)))
   (format (current-error-port) "PLACE-LVM: VG: ~S\n" (obj->name vg))
   (false-if-exception
    (let* ((task (evms 'create (evms 'plugin #:name "LVM2")))
           (lv (find (lambda(obj) (eqv? vg (obj->producer obj)))
                     (task 'acceptable)))
           (size (pr-place-size lv size rest)))
      (format (current-error-port)
              "PLACE-LVM: LV: ~S ~S ~S ~S\n" name size (obj->name lv) (obj->size lv))
      (set! (task 'selected) (list lv))
      (set! (task 'options) `(("name" . ,(pr-lvm-mangle-lv-name name))))
      (and (number? size) (set! (task 'options) `(("size" . ,size))))
      (task 'apply)))))

;;--------------------------------------------------------------
(define (pr-place-raid name size rest)
  (let ((disks (pr-candidates)))
    (and (< 1 (length disks))
         ;; TODO / on raid5, check for separate boot
         (cond ((string=? name "/")
                (pr-raid-place 253 size rest 1 disks))
               (#t
                (or (pr-raid-place 131 size rest 2 disks)
                    (pr-raid-place 131 size rest 1 disks)))))))

(define (pr-raid-place type size rest minp disks)
  (let ((ndisks (length disks)))
    (and (< minp ndisks)
         (let* ((psize (pr-raid-scale-size size minp ndisks))
                (prest (pr-raid-scale-size rest minp ndisks))
                (nplaces (pr-raid-list-places psize disks)))
           (if (any (lambda(p) (null? (cdr p))) nplaces)
               ;; filter out disk(s) without free space for raid part
               (pr-raid-place
                type size rest minp
                (filter-map (lambda(p) (and (not (null? (cdr p))) (car p))) nplaces))
               ;; okay, there are free spaces on remaining disks
               (let ((objs (append-map
                            (lambda(p) (pr-raid-place-part type psize prest p))
                            (map cdr nplaces))))
                 (if (every values objs)
                     (let ((task (evms 'create (pr-raid-plugin minp))))
                       (set! (task 'selected) objs)
                       (task 'apply))
                     (let ()
                       (for-each (lambda(obj) (obj 'delete))
                                 (filter values objs)
                                 #f)))))))))

(define (pr-raid-place-part type psize prest places)
  (and (not (null? places))
       (or (pr-place-plain-one (car places) type psize prest)
           (pr-raid-place-part type psize prest (cdr places)))))

(define (pr-raid-list-places size disks)
  (map
   (lambda(dsk)
     (cons dsk (sort (pr-disk->freespace dsk size) pr-place-size<)))
   disks))

(define (pr-raid-scale-size size minp n)
  (if (= 1 minp) size                   ;raid1
      (pr-scale-size (lambda(s) (round-integer (/ s (- n 1)))) size)))

(define (pr-raid-plugin minp)
  (evms 'plugin #:name
        (if (= 1 minp)
            "MDRaid1RegMgr"
            "MDRaid5RegMgr")))

;---------------------------------------------------------------
(define (pr-disk->freespace disk size)
  (cond  ((null? (obj->parents disk))   ; no partition table yet
          (format #t "DISK->FREESPACE: no pt on ~S, creating\n" (obj->name disk))
          (pr-create-pt disk)
          (pr-disk->freespace disk size))
         ((not (member (obj->segmgr->name disk) '("DosSegMgr" "GptSegMgr")))
          (format #t "DISK->FREESPACE: unknown partition table, skipping\n")
          '())
         (#t
          (let ((numslots (pr-numslots disk)))
            (format #t "DISK->FREESPACE: SLOTS: ~S\n" numslots)
            (let ((parents (obj->parents disk)))
              (fold
               (lambda (seg accum)
                 (if (and (segment-free? seg) (pr-size< size (obj->size seg)))
                     (let-values (((before after) (partition (lambda(segn) (pr-seg< segn seg)) parents)))
                       (format #t "DISK->FREESPACE: ~S X ~S\n"
                               (map obj->name before) (map obj->name (cdr after)))
                       (cond ((< 3 numslots)
                              ;; gpt or empty dos disk: splittable, primary
                              (cons (values seg #t #t) accum))
                             ((not (null? (filter segment-meta? after)))
                              ;; freespace, then ext'd: maybe-splittable, primary
                              (cons (values seg (< 2 numslots) #t) accum))
                             ((< 1 (length (filter segment-meta? before)))
                              ;; ext'd, then free
                              (if (segment-meta? (cadr (reverse before)))
                                  ;; no pri after ext'd: splittable, logical
                                  (cons (values seg #t #f) accum)
                                  ;; pri after ext'd: maybe-splittable, primary
                                  (cons (values seg (< 2 numslots) #f) accum)))
                             (#t
                              ;; primary only
                              (cons (values seg (< 2 numslots) (< 2 numslots)) accum))))
                     accum))
               '() parents))))))

(define (pr-candidates)
  (filter pr-wholedisk?
          (filter pr-candidate?
                  (evms 'object-list #:type 'disk))))

(define (pr-candidate? disk)
  (not (member (obj->name disk)
               (fluid-ref pr-exclude))))

(define (pr-wholedisk? disk)
  (and (pr-purge-fake-volume disk) disk))

(define (pr-purge-fake-volume disk)
  (not (and-let*
        ((vol (assv-ref (disk) 'volume)))
        (or (assv-ref (vol) 'fsim)
            (begin (vol 'delete) #f)))))

(define (pr-numslots disk)
  (let ((segs (filter segment-data? (obj->parents disk))))
    (cond ((string=? "DosSegMgr" (obj->segmgr->name disk))
           (let ((pslots (map (lambda(s) (string-append (obj->name disk) s)) '("1" "2" "3" "4"))))
             (let-values (((pri log) (partition (lambda(x) (member x pslots)) (map obj->name segs))))
               (- 4 (length pri) (if (null? log) 0 1)))))
          ((string=? "GptSegMgr" (obj->segmgr->name disk))
           (- 128 (length segs)))
          (#t 0))))

(define (pr-create-pt disk)
  (let ((task (evms 'assign (evms 'plugin #:name (assumed-segmgr disk)))))
    (set! (task 'selected) (list disk))
    (task 'apply)))

(define (pr-seg< sega segb)
  (< (assv-ref (sega) 'start)
     (assv-ref (segb) 'start)))

(define (pr-size< size num)
  (< (pr-min-size size) num))

(define (pr-scale-size fn size)
  (if (pair? size)
      (cons (fn (car size))
            (let ((max (cdr size)))
              (if (number? max)
                  (fn max)
                  max)))
      (fn size)))

(define (pr-min-size size)
  (let ((min (if (pair? size) (car size) size)))
    (if (number? min) min 0)))

(define (pr-max-size size)
  (if (pair? size) (cdr size) size))
